@@ -18,22 +18,20 @@ templates.
 :note: Directives support added since version 0.6
 """
 
-try:
-    any
-except NameError:
-    from genshi.util import any
 from gettext import NullTranslations
 import os
 import re
+from functools import partial
 from types import FunctionType
 
+import six
+
 from genshi.core import Attrs, Namespace, QName, START, END, TEXT, \
                         XML_NAMESPACE, _ensure, StreamEventKind
-from genshi.template.eval import _ast
 from genshi.template.base import DirectiveFactory, EXPR, SUB, _apply_directives
 from genshi.template.directives import Directive, StripDirective
 from genshi.template.markup import MarkupTemplate, EXEC
-from genshi.compat import IS_PYTHON2
+from genshi.compat import ast, IS_PYTHON2, _ast_Str, _ast_Str_value
 
 __all__ = ['Translator', 'extract']
 __docformat__ = 'restructuredtext en'
@@ -60,21 +58,43 @@ class ExtractableI18NDirective(I18NDirective):
     """Simple interface for directives to support messages extraction."""
 
     def extract(self, translator, stream, gettext_functions=GETTEXT_FUNCTIONS,
-                search_text=True, comment_stack=None):
+                search_text=True, comment_stack=None, context_stack=None):
         raise NotImplementedError
 
 
+contexted = {
+    None: 'pgettext',
+    'gettext': 'pgettext',
+    'ngettext': 'pngettext',
+    'dgettext': 'dpgettext',
+    'dngettext': 'dnpgettext'
+}
+
+
+def contextify(line, func, msg, comment, context):
+    if context:
+        context = context[0]
+        func = contexted.get(func)
+        if func is None:
+            raise ValueError("failure, bogus extraction method")
+        if isinstance(msg, tuple):
+            msg = (context, tuple[0], tuple[1])
+        else:
+            msg = (context, msg)
+    return line, func, msg, comment
+
+
 class CommentDirective(I18NDirective):
     """Implementation of the ``i18n:comment`` template directive which adds
     translation comments.
-    
+
     >>> tmpl = MarkupTemplate('''<html xmlns:i18n="http://genshi.edgewall.org/i18n">
     ...   <p i18n:comment="As in Foo Bar">Foo</p>
     ... </html>''')
     >>> translator = Translator()
     >>> translator.setup(tmpl)
     >>> list(translator.extract(tmpl.stream))
-    [(2, None, u'Foo', [u'As in Foo Bar'])]
+    [(2, None, 'Foo', ['As in Foo Bar'])]
     """
     __slots__ = ['comment']
 
@@ -87,7 +107,7 @@ class CommentDirective(I18NDirective):
 class MsgDirective(ExtractableI18NDirective):
     r"""Implementation of the ``i18n:msg`` directive which marks inner content
     as translatable. Consider the following examples:
-    
+
     >>> tmpl = MarkupTemplate('''<html xmlns:i18n="http://genshi.edgewall.org/i18n">
     ...   <div i18n:msg="">
     ...     <p>Foo</p>
@@ -95,11 +115,11 @@ class MsgDirective(ExtractableI18NDirective):
     ...   </div>
     ...   <p i18n:msg="">Foo <em>bar</em>!</p>
     ... </html>''')
-    
+
     >>> translator = Translator()
     >>> translator.setup(tmpl)
     >>> list(translator.extract(tmpl.stream))
-    [(2, None, u'[1:Foo]\n    [2:Bar]', []), (6, None, u'Foo [1:bar]!', [])]
+    [(2, None, '[1:Foo]\n    [2:Bar]', []), (6, None, 'Foo [1:bar]!', [])]
     >>> print(tmpl.generate().render())
     <html>
       <div><p>Foo</p>
@@ -116,8 +136,8 @@ class MsgDirective(ExtractableI18NDirective):
     ... </html>''')
     >>> translator.setup(tmpl)
     >>> list(translator.extract(tmpl.stream)) #doctest: +NORMALIZE_WHITESPACE
-    [(2, None, u'[1:First Name: %(fname)s]\n    [2:Last Name: %(lname)s]', []),
-    (6, None, u'Foo [1:bar]!', [])]
+    [(2, None, '[1:First Name: %(fname)s]\n    [2:Last Name: %(lname)s]', []),
+    (6, None, 'Foo [1:bar]!', [])]
 
     >>> tmpl = MarkupTemplate('''<html xmlns:i18n="http://genshi.edgewall.org/i18n">
     ...   <div i18n:msg="fname, lname">
@@ -155,20 +175,32 @@ class MsgDirective(ExtractableI18NDirective):
 
     def __call__(self, stream, directives, ctxt, **vars):
         gettext = ctxt.get('_i18n.gettext')
-        if ctxt.get('_i18n.domain'):
+        if ctxt.get('_i18n.domain') and ctxt.get('_i18n.context'):
+            dpgettext = ctxt.get('_i18n.dpgettext')
+            assert hasattr(dpgettext, '__call__'), \
+                'No domain/context gettext function passed'
+            gettext = lambda msg: dpgettext(ctxt.get('_i18n.domain'),
+                                            ctxt.get('_i18n.context'),
+                                            msg)
+        elif ctxt.get('_i18n.domain'):
             dgettext = ctxt.get('_i18n.dgettext')
             assert hasattr(dgettext, '__call__'), \
                 'No domain gettext function passed'
             gettext = lambda msg: dgettext(ctxt.get('_i18n.domain'), msg)
+        elif ctxt.get('_i18n.context'):
+            pgettext = ctxt.get('_i18n.pgettext')
+            assert hasattr(pgettext, '__call__'), \
+                'No context gettext function passed'
+            gettext = lambda msg: pgettext(ctxt.get('_i18n.context'), msg)
 
         def _generate():
             msgbuf = MessageBuffer(self)
-            previous = stream.next()
+            previous = next(stream)
             if previous[0] is START:
                 yield previous
             else:
                 msgbuf.append(*previous)
-            previous = stream.next()
+            previous = next(stream)
             for kind, data, pos in stream:
                 msgbuf.append(*previous)
                 previous = kind, data, pos
@@ -183,18 +215,18 @@ class MsgDirective(ExtractableI18NDirective):
         return _apply_directives(_generate(), directives, ctxt, vars)
 
     def extract(self, translator, stream, gettext_functions=GETTEXT_FUNCTIONS,
-                search_text=True, comment_stack=None):
+                search_text=True, comment_stack=None, context_stack=None):
         msgbuf = MessageBuffer(self)
         strip = False
 
         stream = iter(stream)
-        previous = stream.next()
+        previous = next(stream)
         if previous[0] is START:
             for message in translator._extract_attrs(previous,
                                                      gettext_functions,
                                                      search_text=search_text):
                 yield message
-            previous = stream.next()
+            previous = next(stream)
             strip = True
         for event in stream:
             if event[0] is START:
@@ -207,7 +239,8 @@ class MsgDirective(ExtractableI18NDirective):
         if not strip:
             msgbuf.append(*previous)
 
-        yield self.lineno, None, msgbuf.format(), comment_stack[-1:]
+        yield contextify(
+            self.lineno, None, msgbuf.format(), comment_stack[-1:], context_stack[-1:])
 
 
 class ChooseBranchDirective(I18NDirective):
@@ -218,14 +251,14 @@ class ChooseBranchDirective(I18NDirective):
         msgbuf = MessageBuffer(self)
         stream = _apply_directives(stream, directives, ctxt, vars)
 
-        previous = stream.next()
+        previous = next(stream)
         if previous[0] is START:
             yield previous
         else:
             msgbuf.append(*previous)
 
         try:
-            previous = stream.next()
+            previous = next(stream)
         except StopIteration:
             # For example <i18n:singular> or <i18n:plural> directives
             yield MSGBUF, (), -1 # the place holder for msgbuf output
@@ -244,9 +277,10 @@ class ChooseBranchDirective(I18NDirective):
         ctxt['_i18n.choose.%s' % self.tagname] = msgbuf
 
     def extract(self, translator, stream, gettext_functions=GETTEXT_FUNCTIONS,
-                search_text=True, comment_stack=None, msgbuf=None):
+                search_text=True, comment_stack=None, context_stack=None,
+                msgbuf=None):
         stream = iter(stream)
-        previous = stream.next()
+        previous = next(stream)
 
         if previous[0] is START:
             # skip the enclosing element
@@ -254,7 +288,7 @@ class ChooseBranchDirective(I18NDirective):
                                                      gettext_functions,
                                                      search_text=search_text):
                 yield message
-            previous = stream.next()
+            previous = next(stream)
 
         for event in stream:
             if previous[0] is START:
@@ -282,13 +316,13 @@ class PluralDirective(ChooseBranchDirective):
 class ChooseDirective(ExtractableI18NDirective):
     """Implementation of the ``i18n:choose`` directive which provides plural
     internationalisation of strings.
-    
+
     This directive requires at least one parameter, the one which evaluates to
     an integer which will allow to choose the plural/singular form. If you also
     have expressions inside the singular and plural version of the string you
     also need to pass a name for those parameters. Consider the following
     examples:
-    
+
     >>> tmpl = MarkupTemplate('''<html xmlns:i18n="http://genshi.edgewall.org/i18n">
     ...   <div i18n:choose="num; num">
     ...     <p i18n:singular="">There is $num coin</p>
@@ -298,8 +332,8 @@ class ChooseDirective(ExtractableI18NDirective):
     >>> translator = Translator()
     >>> translator.setup(tmpl)
     >>> list(translator.extract(tmpl.stream)) #doctest: +NORMALIZE_WHITESPACE
-    [(2, 'ngettext', (u'There is %(num)s coin',
-                      u'There are %(num)s coins'), [])]
+    [(2, 'ngettext', ('There is %(num)s coin',
+                      'There are %(num)s coins'), [])]
 
     >>> tmpl = MarkupTemplate('''<html xmlns:i18n="http://genshi.edgewall.org/i18n">
     ...   <div i18n:choose="num; num">
@@ -331,8 +365,8 @@ class ChooseDirective(ExtractableI18NDirective):
     ... </html>''')
     >>> translator.setup(tmpl)
     >>> list(translator.extract(tmpl.stream)) #doctest: +NORMALIZE_WHITESPACE
-    [(2, 'ngettext', (u'There is %(num)s coin',
-                      u'There are %(num)s coins'), [])]
+    [(2, 'ngettext', ('There is %(num)s coin',
+                      'There are %(num)s coins'), [])]
     """
     __slots__ = ['numeral', 'params', 'lineno']
 
@@ -349,7 +383,7 @@ class ChooseDirective(ExtractableI18NDirective):
     def attach(cls, template, stream, value, namespaces, pos):
         if type(value) is dict:
             numeral = value.get('numeral', '').strip()
-            assert numeral is not '', "at least pass the numeral param"
+            assert numeral != '', "at least pass the numeral param"
             params = [v.strip() for v in value.get('params', '').split(',')]
             value = '%s; ' % numeral + ', '.join(params)
         return super(ChooseDirective, cls).attach(template, stream, value,
@@ -362,9 +396,15 @@ class ChooseDirective(ExtractableI18NDirective):
 
         ngettext = ctxt.get('_i18n.ngettext')
         assert hasattr(ngettext, '__call__'), 'No ngettext function available'
+        npgettext = ctxt.get('_i18n.npgettext')
+        if not npgettext:
+            npgettext = lambda c, s, p, n: ngettext(s, p, n)
         dngettext = ctxt.get('_i18n.dngettext')
         if not dngettext:
             dngettext = lambda d, s, p, n: ngettext(s, p, n)
+        dnpgettext = ctxt.get('_i18n.dnpgettext')
+        if not dnpgettext:
+            dnpgettext = lambda d, c, s, p, n: dngettext(d, s, p, n)
 
         new_stream = []
         singular_stream = None
@@ -395,7 +435,14 @@ class ChooseDirective(ExtractableI18NDirective):
             else:
                 new_stream.append(event)
 
-        if ctxt.get('_i18n.domain'):
+        if ctxt.get('_i18n.context') and ctxt.get('_i18n.domain'):
+            ngettext = lambda s, p, n: dnpgettext(ctxt.get('_i18n.domain'),
+                                                  ctxt.get('_i18n.context'),
+                                                  s, p, n)
+        elif ctxt.get('_i18n.context'):
+            ngettext = lambda s, p, n: npgettext(ctxt.get('_i18n.context'),
+                                                 s, p, n)
+        elif ctxt.get('_i18n.domain'):
             ngettext = lambda s, p, n: dngettext(ctxt.get('_i18n.domain'),
                                                  s, p, n)
 
@@ -424,10 +471,10 @@ class ChooseDirective(ExtractableI18NDirective):
         ctxt.pop()
 
     def extract(self, translator, stream, gettext_functions=GETTEXT_FUNCTIONS,
-                search_text=True, comment_stack=None):
+                search_text=True, comment_stack=None, context_stack=None):
         strip = False
         stream = iter(stream)
-        previous = stream.next()
+        previous = next(stream)
 
         if previous[0] is START:
             # skip the enclosing element
@@ -435,7 +482,7 @@ class ChooseDirective(ExtractableI18NDirective):
                                                      gettext_functions,
                                                      search_text=search_text):
                 yield message
-            previous = stream.next()
+            previous = next(stream)
             strip = True
 
         singular_msgbuf = MessageBuffer(self)
@@ -448,12 +495,12 @@ class ChooseDirective(ExtractableI18NDirective):
                     if isinstance(directive, SingularDirective):
                         for message in directive.extract(translator,
                                 substream, gettext_functions, search_text,
-                                comment_stack, msgbuf=singular_msgbuf):
+                                comment_stack, context_stack, msgbuf=singular_msgbuf):
                             yield message
                     elif isinstance(directive, PluralDirective):
                         for message in directive.extract(translator,
                                 substream, gettext_functions, search_text,
-                                comment_stack, msgbuf=plural_msgbuf):
+                                comment_stack, context_stack, msgbuf=plural_msgbuf):
                             yield message
                     elif not isinstance(directive, StripDirective):
                         singular_msgbuf.append(*previous)
@@ -472,9 +519,9 @@ class ChooseDirective(ExtractableI18NDirective):
             singular_msgbuf.append(*previous)
             plural_msgbuf.append(*previous)
 
-        yield self.lineno, 'ngettext', \
+        yield contextify(self.lineno, 'ngettext', \
             (singular_msgbuf.format(), plural_msgbuf.format()), \
-            comment_stack[-1:]
+            comment_stack[-1:], context_stack[-1:])
 
     def _is_plural(self, numeral, ngettext):
         # XXX: should we test which form was chosen like this!?!?!?
@@ -488,7 +535,7 @@ class ChooseDirective(ExtractableI18NDirective):
 class DomainDirective(I18NDirective):
     """Implementation of the ``i18n:domain`` directive which allows choosing
     another i18n domain(catalog) to translate from.
-    
+
     >>> from genshi.filters.tests.i18n import DummyTranslations
     >>> tmpl = MarkupTemplate('''<html xmlns:i18n="http://genshi.edgewall.org/i18n">
     ...   <p i18n:msg="">Bar</p>
@@ -540,12 +587,34 @@ class DomainDirective(I18NDirective):
         ctxt.pop()
 
 
+class ContextDirective(I18NDirective):
+    __slots__ = ['context']
+
+    def __init__(self, value, template=None, namespaces=None, lineno=-1,
+                 offset=-1):
+        Directive.__init__(self, None, template, namespaces, lineno, offset)
+        self.context = value
+
+    @classmethod
+    def attach(cls, template, stream, value, namespaces, pos):
+        if type(value) is dict:
+            value = value.get('name')
+        return super(ContextDirective, cls).attach(template, stream, value,
+                                                   namespaces, pos)
+
+    def __call__(self, stream, directives, ctxt, **vars):
+        ctxt.push({'_i18n.context': self.context})
+        for event in _apply_directives(stream, directives, ctxt, vars):
+            yield event
+        ctxt.pop()
+
+
 class Translator(DirectiveFactory):
     """Can extract and translate localizable strings from markup streams and
     templates.
-    
+
     For example, assume the following template:
-    
+
     >>> tmpl = MarkupTemplate('''<html xmlns:py="http://genshi.edgewall.org/">
     ...   <head>
     ...     <title>Example</title>
@@ -555,25 +624,25 @@ class Translator(DirectiveFactory):
     ...     <p>${_("Hello, %(name)s") % dict(name=username)}</p>
     ...   </body>
     ... </html>''', filename='example.html')
-    
+
     For demonstration, we define a dummy ``gettext``-style function with a
     hard-coded translation table, and pass that to the `Translator` initializer:
-    
+
     >>> def pseudo_gettext(string):
     ...     return {
     ...         'Example': 'Beispiel',
     ...         'Hello, %(name)s': 'Hallo, %(name)s'
     ...     }[string]
     >>> translator = Translator(pseudo_gettext)
-    
+
     Next, the translator needs to be prepended to any already defined filters
     on the template:
-    
+
     >>> tmpl.filters.insert(0, translator)
-    
+
     When generating the template output, our hard-coded translations should be
     applied as expected:
-    
+
     >>> print(tmpl.generate(username='Hans', _=pseudo_gettext))
     <html>
       <head>
@@ -584,7 +653,7 @@ class Translator(DirectiveFactory):
         <p>Hallo, Hans</p>
       </body>
     </html>
-    
+
     Note that elements defining ``xml:lang`` attributes that do not contain
     variable expressions are ignored by this filter. That can be used to
     exclude specific parts of a template from being extracted and translated.
@@ -593,6 +662,7 @@ class Translator(DirectiveFactory):
     directives = [
         ('domain', DomainDirective),
         ('comment', CommentDirective),
+        ('ctxt', ContextDirective),
         ('msg', MsgDirective),
         ('choose', ChooseDirective),
         ('singular', SingularDirective),
@@ -612,7 +682,7 @@ class Translator(DirectiveFactory):
     def __init__(self, translate=NullTranslations(), ignore_tags=IGNORE_TAGS,
                  include_attrs=INCLUDE_ATTRS, extract_text=True):
         """Initialize the translator.
-        
+
         :param translate: the translation function, for example ``gettext`` or
                           ``ugettext``.
         :param ignore_tags: a set of tag names that should not be localized
@@ -620,7 +690,7 @@ class Translator(DirectiveFactory):
         :param extract_text: whether the content of text nodes should be
                              extracted, or only text in explicit ``gettext``
                              function calls
-        
+
         :note: Changed in 0.6: the `translate` parameter can now be either
                a ``gettext``-style function, or an object compatible with the
                ``NullTransalations`` or ``GNUTranslations`` interface
@@ -633,13 +703,13 @@ class Translator(DirectiveFactory):
     def __call__(self, stream, ctxt=None, translate_text=True,
                  translate_attrs=True):
         """Translate any localizable strings in the given stream.
-        
+
         This function shouldn't be called directly. Instead, an instance of
         the `Translator` class should be registered as a filter with the
         `Template` or the `TemplateLoader`, or applied as a regular stream
         filter. If used as a template filter, it should be inserted in front of
         all the default filters.
-        
+
         :param stream: the markup event stream
         :param ctxt: the template context (not used)
         :param translate_text: whether text nodes should be translated (used
@@ -677,16 +747,44 @@ class Translator(DirectiveFactory):
             except AttributeError:
                 dgettext = lambda _, y: gettext(y)
                 dngettext = lambda _, s, p, n: ngettext(s, p, n)
+            try:
+                if IS_PYTHON2:
+                    pgettext = self.translate.upgettext
+                    dpgettext = self.translate.dupgettext
+                    npgettext = self.translate.unpgettext
+                    dnpgettext = self.translate.dunpgettext
+                else:
+                    pgettext = self.translate.pgettext
+                    dpgettext = self.translate.dpgettext
+                    npgettext = self.translate.npgettext
+                    dnpgettext = self.translate.dnpgettext
+            except AttributeError:
+                pgettext = lambda _, y: gettext(y)
+                dpgettext = lambda d, _, y: dgettext(d, y)
+                npgettext = lambda _, s, p, n: ngettext(s, p, n)
+                dnpgettext = lambda d, _, s, p, n: dngettext(d, s, p, n)
             if ctxt:
                 ctxt['_i18n.gettext'] = gettext
                 ctxt['_i18n.ngettext'] = ngettext
                 ctxt['_i18n.dgettext'] = dgettext
                 ctxt['_i18n.dngettext'] = dngettext
+                ctxt['_i18n.pgettext'] = pgettext
+                ctxt['_i18n.npgettext'] = npgettext
+                ctxt['_i18n.dpgettext'] = dpgettext
+                ctxt['_i18n.dnpgettext'] = dnpgettext
 
         if ctxt and ctxt.get('_i18n.domain'):
             # TODO: This can cause infinite recursion if dgettext is defined
             #       via the AttributeError case above!
-            gettext = lambda msg: dgettext(ctxt.get('_i18n.domain'), msg)
+            gettext = partial(dgettext, ctxt.get('_i18n.domain'))
+
+        if ctxt and ctxt.get('_i18n.context'):
+            if getattr(gettext, 'func', None):
+                gettext = partial(dpgettext,
+                                  ctxt['_i18n.domain'],
+                                  ctxt['_i18n.context'])
+            else:
+                gettext = partial(pgettext, ctxt['_i18n.context'])
 
         for kind, data, pos in stream:
 
@@ -703,7 +801,7 @@ class Translator(DirectiveFactory):
             if kind is START:
                 tag, attrs = data
                 if tag in self.ignore_tags or \
-                        isinstance(attrs.get(xml_lang), basestring):
+                        isinstance(attrs.get(xml_lang), six.string_types):
                     skip += 1
                     yield kind, data, pos
                     continue
@@ -713,9 +811,10 @@ class Translator(DirectiveFactory):
 
                 for name, value in attrs:
                     newval = value
-                    if isinstance(value, basestring):
-                        if translate_attrs and name in include_attrs:
-                            newval = gettext(value)
+                    if isinstance(value, six.string_types):
+                        text = value.strip()
+                        if translate_attrs and name in include_attrs and text:
+                            newval = gettext(text)
                     else:
                         newval = list(
                             self(_ensure(value), ctxt, translate_text=False)
@@ -732,12 +831,13 @@ class Translator(DirectiveFactory):
             elif translate_text and kind is TEXT:
                 text = data.strip()
                 if text:
-                    data = data.replace(text, unicode(gettext(text)))
+                    data = data.replace(text, six.text_type(gettext(text)))
                 yield kind, data, pos
 
             elif kind is SUB:
                 directives, substream = data
                 current_domain = None
+                current_context = None
                 for idx, directive in enumerate(directives):
                     # Organize directives to make everything work
                     # FIXME: There's got to be a better way to do this!
@@ -748,6 +848,15 @@ class Translator(DirectiveFactory):
                         # Put domain directive as the first one in order to
                         # update context before any other directives evaluation
                         directives.insert(0, directives.pop(idx))
+                    if isinstance(directive, ContextDirective):
+                        # Grab current (msg)context and update context
+                        current_context = directive.context
+                        ctxt.push({'_i18n.context': current_context})
+                        # Put context directive either first in the case of
+                        # no domain, or 2nd in the case there is a domain, to
+                        # update context before any other directives evaluation
+                        directives.insert(1 if current_domain else 0,
+                                          directives.pop(idx))
 
                 # If this is an i18n directive, no need to translate text
                 # nodes here
@@ -755,6 +864,7 @@ class Translator(DirectiveFactory):
                     isinstance(d, ExtractableI18NDirective)
                     for d in directives
                 ])
+
                 substream = list(self(substream, ctxt,
                                       translate_text=not is_i18n_directive,
                                       translate_attrs=translate_attrs))
@@ -762,16 +872,18 @@ class Translator(DirectiveFactory):
 
                 if current_domain:
                     ctxt.pop()
+                if current_context:
+                    ctxt.pop()
             else:
                 yield kind, data, pos
 
     def extract(self, stream, gettext_functions=GETTEXT_FUNCTIONS,
-                search_text=True, comment_stack=None):
+                search_text=True, comment_stack=None, context_stack=None):
         """Extract localizable strings from the given template stream.
-        
+
         For every string found, this function yields a ``(lineno, function,
         message, comments)`` tuple, where:
-        
+
         * ``lineno`` is the number of the line on which the string was found,
         * ``function`` is the name of the ``gettext`` function used (if the
           string was extracted from embedded Python code), and
@@ -780,7 +892,7 @@ class Translator(DirectiveFactory):
            arguments).
         *  ``comments`` is a list of comments related to the message, extracted
            from ``i18n:comment`` attributes found in the markup
-        
+
         >>> tmpl = MarkupTemplate('''<html xmlns:py="http://genshi.edgewall.org/">
         ...   <head>
         ...     <title>Example</title>
@@ -793,10 +905,10 @@ class Translator(DirectiveFactory):
         ... </html>''', filename='example.html')
         >>> for line, func, msg, comments in Translator().extract(tmpl.stream):
         ...    print('%d, %r, %r' % (line, func, msg))
-        3, None, u'Example'
-        6, None, u'Example'
-        7, '_', u'Hello, %(name)s'
-        8, 'ngettext', (u'You have %d item', u'You have %d items', None)
+        3, None, 'Example'
+        6, None, 'Example'
+        7, '_', 'Hello, %(name)s'
+        8, 'ngettext', ('You have %d item', 'You have %d items', None)
         
         :param stream: the event stream to extract strings from; can be a
                        regular stream or a template stream
@@ -805,7 +917,7 @@ class Translator(DirectiveFactory):
                                   functions
         :param search_text: whether the content of text nodes should be
                             extracted (used internally)
-        
+
         :note: Changed in 0.4.1: For a function with multiple string arguments
                (such as ``ngettext``), a single item with a tuple of strings is
                yielded, instead an item for each string argument.
@@ -816,6 +928,8 @@ class Translator(DirectiveFactory):
             search_text = False
         if comment_stack is None:
             comment_stack = []
+        if context_stack is None:
+            context_stack = []
         skip = 0
 
         xml_lang = XML_NAMESPACE['lang']
@@ -830,7 +944,7 @@ class Translator(DirectiveFactory):
             if kind is START and not skip:
                 tag, attrs = data
                 if tag in self.ignore_tags or \
-                        isinstance(attrs.get(xml_lang), basestring):
+                        isinstance(attrs.get(xml_lang), six.string_types):
                     skip += 1
                     continue
 
@@ -842,7 +956,8 @@ class Translator(DirectiveFactory):
             elif not skip and search_text and kind is TEXT:
                 text = data.strip()
                 if text and [ch for ch in text if ch.isalpha()]:
-                    yield pos[1], None, text, comment_stack[-1:]
+                    yield contextify(pos[1], None, text, comment_stack[-1:],
+                                     context_stack[-1:])
 
             elif kind is EXPR or kind is EXEC:
                 for funcname, strings in extract_from_code(data,
@@ -853,6 +968,7 @@ class Translator(DirectiveFactory):
             elif kind is SUB:
                 directives, substream = data
                 in_comment = False
+                in_context = False
 
                 for idx, directive in enumerate(directives):
                     # Do a first loop to see if there's a comment directive
@@ -866,14 +982,26 @@ class Translator(DirectiveFactory):
                             for message in self.extract(
                                     substream, gettext_functions,
                                     search_text=search_text and not skip,
-                                    comment_stack=comment_stack):
+                                    comment_stack=comment_stack,
+                                    context_stack=context_stack):
+                                yield message
+                        directives.pop(idx)
+                    elif isinstance(directive, ContextDirective):
+                        in_context = True
+                        context_stack.append(directive.context)
+                        if len(directives) == 1:
+                            for message in self.extract(
+                                    substream, gettext_functions,
+                                    search_text=search_text and not skip,
+                                    comment_stack=comment_stack,
+                                    context_stack=context_stack):
                                 yield message
                         directives.pop(idx)
                     elif not isinstance(directive, I18NDirective):
                         # Remove all other non i18n directives from the process
                         directives.pop(idx)
 
-                if not directives and not in_comment:
+                if not directives and not in_comment and not in_context:
                     # Extract content if there's no directives because
                     # strip was pop'ed and not because comment was pop'ed.
                     # Extraction in this case has been taken care of.
@@ -887,18 +1015,23 @@ class Translator(DirectiveFactory):
                         for message in directive.extract(self,
                                 substream, gettext_functions,
                                 search_text=search_text and not skip,
-                                comment_stack=comment_stack):
+                                comment_stack=comment_stack,
+                                context_stack=context_stack):
                             yield message
                     else:
                         for message in self.extract(
                                 substream, gettext_functions,
                                 search_text=search_text and not skip,
-                                comment_stack=comment_stack):
+                                comment_stack=comment_stack,
+                                context_stack=context_stack):
                             yield message
 
                 if in_comment:
                     comment_stack.pop()
 
+                if in_context:
+                    context_stack.pop()
+
     def get_directive_index(self, dir_cls):
         total = len(self._dir_order)
         if dir_cls in self._dir_order:
@@ -908,7 +1041,7 @@ class Translator(DirectiveFactory):
     def setup(self, template):
         """Convenience function to register the `Translator` filter and the
         related directives with the given template.
-        
+
         :param template: a `Template` instance
         """
         template.filters.insert(0, self)
@@ -917,7 +1050,7 @@ class Translator(DirectiveFactory):
 
     def _extract_attrs(self, event, gettext_functions, search_text):
         for name, value in event[1][1]:
-            if search_text and isinstance(value, basestring):
+            if search_text and isinstance(value, six.string_types):
                 if name in self.include_attrs:
                     text = value.strip()
                     if text:
@@ -930,13 +1063,13 @@ class Translator(DirectiveFactory):
 
 class MessageBuffer(object):
     """Helper class for managing internationalized mixed content.
-    
+
     :since: version 0.5
     """
 
     def __init__(self, directive=None):
         """Initialize the message buffer.
-        
+
         :param directive: the directive owning the buffer
         :type directive: I18NDirective
         """
@@ -963,7 +1096,7 @@ class MessageBuffer(object):
 
     def append(self, kind, data, pos):
         """Append a stream event to the buffer.
-        
+
         :param kind: the stream event kind
         :param data: the event data
         :param pos: the position of the event in the source
@@ -995,7 +1128,7 @@ class MessageBuffer(object):
                     params = "(%s)" % params
                 raise IndexError("%d parameters%s given to 'i18n:%s' but "
                                  "%d or more expressions used in '%s', line %s"
-                                 % (len(self.orig_params), params, 
+                                 % (len(self.orig_params), params,
                                     self.directive.tagname,
                                     len(self.orig_params) + 1,
                                     os.path.basename(pos[0] or
@@ -1005,7 +1138,7 @@ class MessageBuffer(object):
             self._add_event(self.stack[-1], (kind, data, pos))
             self.values[param] = (kind, data, pos)
         else:
-            if kind is START: 
+            if kind is START:
                 self.string.append('[%d:' % self.order)
                 self.stack.append(self.order)
                 self._add_event(self.stack[-1], (kind, data, pos))
@@ -1027,7 +1160,7 @@ class MessageBuffer(object):
     def translate(self, string, regex=re.compile(r'%\((\w+)\)s')):
         """Interpolate the given message translation with the events in the
         buffer and return the translated stream.
-        
+
         :param string: the translated message string
         """
         substream = None
@@ -1122,19 +1255,19 @@ class MessageBuffer(object):
 def parse_msg(string, regex=re.compile(r'(?:\[(\d+)\:)|(?<!\\)\]')):
     """Parse a translated message using Genshi mixed content message
     formatting.
-    
+
     >>> parse_msg("See [1:Help].")
     [(0, 'See '), (1, 'Help'), (0, '.')]
-    
+
     >>> parse_msg("See [1:our [2:Help] page] for details.")
     [(0, 'See '), (1, 'our '), (2, 'Help'), (1, ' page'), (0, ' for details.')]
-    
+
     >>> parse_msg("[2:Details] finden Sie in [1:Hilfe].")
     [(2, 'Details'), (0, ' finden Sie in '), (1, 'Hilfe'), (0, '.')]
-    
+
     >>> parse_msg("[1:] Bilder pro Seite anzeigen.")
     [(1, ''), (0, ' Bilder pro Seite anzeigen.')]
-    
+
     :param string: the translated message string
     :return: a list of ``(order, string)`` tuples
     :rtype: `list`
@@ -1166,16 +1299,16 @@ def parse_msg(string, regex=re.compile(r'(?:\[(\d+)\:)|(?<!\\)\]')):
 
 def extract_from_code(code, gettext_functions):
     """Extract strings from Python bytecode.
-    
+
     >>> from genshi.template.eval import Expression
     >>> expr = Expression('_("Hello")')
     >>> list(extract_from_code(expr, GETTEXT_FUNCTIONS))
-    [('_', u'Hello')]
+    [('_', 'Hello')]
     
     >>> expr = Expression('ngettext("You have %(num)s item", '
     ...                            '"You have %(num)s items", num)')
     >>> list(extract_from_code(expr, GETTEXT_FUNCTIONS))
-    [('ngettext', (u'You have %(num)s item', u'You have %(num)s items', None))]
+    [('ngettext', ('You have %(num)s item', 'You have %(num)s items', None))]
     
     :param code: the `Code` object
     :type code: `genshi.template.eval.Code`
@@ -1183,14 +1316,15 @@ def extract_from_code(code, gettext_functions):
     :since: version 0.5
     """
     def _walk(node):
-        if isinstance(node, _ast.Call) and isinstance(node.func, _ast.Name) \
+        if isinstance(node, ast.Call) and isinstance(node.func, ast.Name) \
                 and node.func.id in gettext_functions:
             strings = []
             def _add(arg):
-                if isinstance(arg, _ast.Str) and isinstance(arg.s, unicode):
-                    strings.append(arg.s)
-                elif isinstance(arg, _ast.Str):
-                    strings.append(unicode(arg.s, 'utf-8'))
+                if isinstance(arg, _ast_Str) \
+                        and isinstance(_ast_Str_value(arg), six.text_type):
+                    strings.append(_ast_Str_value(arg))
+                elif isinstance(arg, _ast_Str):
+                    strings.append(six.text_type(_ast_Str_value(arg), 'utf-8'))
                 elif arg:
                     strings.append(None)
             [_add(arg) for arg in node.args]
@@ -1210,7 +1344,7 @@ def extract_from_code(code, gettext_functions):
                 if isinstance(child, list):
                     for elem in child:
                         children.append(elem)
-                elif isinstance(child, _ast.AST):
+                elif isinstance(child, ast.AST):
                     children.append(child)
             for child in children:
                 for funcname, strings in _walk(child):
@@ -1220,7 +1354,7 @@ def extract_from_code(code, gettext_functions):
 
 def extract(fileobj, keywords, comment_tags, options):
     """Babel extraction method for Genshi templates.
-    
+
     :param fileobj: the file-like object the messages should be extracted from
     :param keywords: a list of keywords (i.e. function names) that should be
                      recognized as translation functions
@@ -1231,22 +1365,22 @@ def extract(fileobj, keywords, comment_tags, options):
     :rtype: ``iterator``
     """
     template_class = options.get('template_class', MarkupTemplate)
-    if isinstance(template_class, basestring):
+    if isinstance(template_class, six.string_types):
         module, clsname = template_class.split(':', 1)
         template_class = getattr(__import__(module, {}, {}, [clsname]), clsname)
     encoding = options.get('encoding', None)
 
     extract_text = options.get('extract_text', True)
-    if isinstance(extract_text, basestring):
+    if isinstance(extract_text, six.string_types):
         extract_text = extract_text.lower() in ('1', 'on', 'yes', 'true')
 
     ignore_tags = options.get('ignore_tags', Translator.IGNORE_TAGS)
-    if isinstance(ignore_tags, basestring):
+    if isinstance(ignore_tags, six.string_types):
         ignore_tags = ignore_tags.split()
     ignore_tags = [QName(tag) for tag in ignore_tags]
 
     include_attrs = options.get('include_attrs', Translator.INCLUDE_ATTRS)
-    if isinstance(include_attrs, basestring):
+    if isinstance(include_attrs, six.string_types):
         include_attrs = include_attrs.split()
     include_attrs = [QName(attr) for attr in include_attrs]
 
